import collections
import getopt
import numpy
import os
import random
import sys
import unittest
from time import process_time as clock

from amuse.community.ph4.interface import ph4 as grav
from amuse.community.smalln.interface import SmallN
from amuse.community.kepler.interface import Kepler
from amuse.couple import multiples

from amuse.units import nbody_system
from amuse.units import units
from amuse.units import quantities

from amuse import datamodel
from amuse.datamodel import particle_attributes as pa
from amuse.rfi.core import is_mpd_running
from amuse.ic.plummer import new_plummer_model
from amuse.ic.salpeter import new_salpeter_mass_distribution_nbody

def print_log(pre, time, gravity, E0 = 0.0 | nbody_system.energy, cpu0 = 0.0):
    cpu = clock()
    N = len(gravity.particles)
    M = gravity.total_mass
    U = gravity.potential_energy
    T = gravity.kinetic_energy
    Etop = T + U
    Nmul, Nbin, Emul = gravity.get_total_multiple_energy()
    tmp1,tmp2,Emul2 = gravity.get_total_multiple_energy2()
    Etot = Etop + Emul
    Eext = gravity.multiples_external_tidal_correction
    Eint = gravity.multiples_internal_tidal_correction
    Eerr = gravity.multiples_integration_energy_error
    Edel = gravity.multiples_external_tidal_correction \
		+ gravity.multiples_internal_tidal_correction \
		+ gravity.multiples_integration_energy_error
    Ecor = Etot - Edel
    if E0 == 0 | nbody_system.energy: E0 = Ecor
    Rvir = -0.5*M*M/U
    Q = -T/U
    com = pa.center_of_mass(gravity.particles)
    comv = pa.center_of_mass_velocity(gravity.particles)
    dcen,rcore,rhocore = pa.densitycentre_coreradius_coredens(gravity.particles)
    cmx,cmy,cmz = dcen
    lagr,mf = pa.LagrangianRadii(gravity.particles, cm=dcen)  # no units!

    print('')
    print(pre+"time=", time.number)
    print(pre+"CPU=", cpu - cpu0)
    print(pre+"Ntot=", N)
    print(pre+"mass=", M.number)
    print(pre+"Etot=", Etot.number)
    print(pre+"Etop=", Etop.number)
    print(pre+"Eext=", Eext.number)
    print(pre+"Eint=", Eint.number)
    print(pre+"Eerr=", Eerr.number)
    print(pre+"Edel=", Edel.number)
    print(pre+"Ecor=", Ecor.number)
    print(pre+"dE/E=", Ecor/E0 - 1)
    print(pre+"Rvir=", Rvir.number)
    print(pre+"Qvir=", Q)
    cmx,cmy,cmz = com
    print(pre+"cmpos[3]= %.8f %.8f %.8f" % (cmx.number, cmy.number, cmz.number))
    cmx,cmy,cmz = comv
    print(pre+"cmvel[3]= %.8f %.8f %.8f" % (cmx.number, cmy.number, cmz.number))
    cmx,cmy,cmz = dcen
    print(pre+"dcpos[3]= %.8f %.8f %.8f" % (cmx.number, cmy.number, cmz.number))
    print(pre+"Rcore=", rcore.number)
    print(pre+"Mcore=", (rhocore*rcore**3).number)	# fake...
    print(pre+"Mlagr[9]=", end=' ')
    for m in mf: print("%.4f" % (m), end=' ')
    print('')
    print(pre+"Rlagr[9]=", end=' ')
    for r in lagr.number: print("%.8f" % (r), end=' ')
    print('')
    kT = T/N
    Nmul,Nbin,Emul = gravity.print_multiples2(pre, kT, dcen)
    print(pre+"Nmul=", Nmul)
    print(pre+"Nbin=", Nbin)
    print(pre+"Emul= %.5f" % (Emul.number))
    print(pre+"Emul2= %.5f" % (Emul2.number))
    print(pre+"Emul/kT= %.5f" % (Emul.number/kT.number))
    print(pre+"Emul/E= %.5f" % (Emul.number/Etot.number))
    print('')

    sys.stdout.flush()
    return Ecor,cpu

SMALLN = None
def new_smalln():
    SMALLN.reset()
    return SMALLN
    
def init_smalln():
    global SMALLN
    sys.stdout.flush()
    SMALLN = SmallN()
    sys.stdout.flush()
    SMALLN.parameters.timestep_parameter = 0.1
    #SMALLN.parameters.cm_index = 2001		# don't set this here!!
    sys.stdout.flush()

def init_kepler(star1, star2):        
    try:
        star1.mass.value_in(units.kg) # see if SI units, throw exception if not
        unit_converter \
            = nbody_system.nbody_to_si(star1.mass + star2.mass,
                                       (star2.position-star1.position).length())
    except Exception as ex:
        unit_converter = None
        
    kep = Kepler(unit_converter, redirection = "none")
    kep.initialize_code()

    return kep

def run_ph4(infile = None, outfile = None,
            number_of_stars = 100, number_of_binaries = 0,
            end_time = 10 | nbody_system.time,
            delta_t = 1 | nbody_system.time,
            n_workers = 1, use_gpu = 1, gpu_worker = 1,
            salpeter = 0,
            accuracy_parameter = 0.1,
            softening_length = 0.0 | nbody_system.length,
            manage_encounters = 1, random_seed = 1234,
            debug_level = 1):

    if random_seed <= 0:
        numpy.random.seed()
        random_seed = numpy.random.randint(1, pow(2,31)-1)
    numpy.random.seed(random_seed)
    print("random seed =", random_seed)

    if infile != None: print("input file =", infile)
    print("end_time =", end_time)
    print("delta_t =", delta_t)
    print("n_workers =", n_workers)
    print("use_gpu =", use_gpu)
    print("manage_encounters =", manage_encounters)
    print("\ninitializing the gravity module")
    sys.stdout.flush()

    init_smalln()
    
    # Note that there are actually three GPU options:
    #
    #	1. use the GPU code and allow GPU use (default)
    #	2. use the GPU code but disable GPU use (-g)
    #	3. use the non-GPU code (-G)

    if gpu_worker == 1:
        try:
            gravity = grav(number_of_workers = n_workers,
                           redirection = "none", mode = "gpu")
        except Exception as ex:
            gravity = grav(number_of_workers = n_workers,
                           redirection = "none")
    else:
        gravity = grav(number_of_workers = n_workers,
                       redirection = "none")

    gravity.initialize_code()
    gravity.parameters.set_defaults()

    #-----------------------------------------------------------------

    if infile == None:

        print("making a Plummer model")
        stars = new_plummer_model(number_of_stars)

        id = numpy.arange(number_of_stars)
        stars.id = id+1

        print("setting particle masses and radii")
        if salpeter == 0:
            print('equal masses')
            total_mass = 1.0 | nbody_system.mass
            scaled_mass = total_mass / number_of_stars
        else:
            print('salpeter mass function')
            scaled_mass = new_salpeter_mass_distribution_nbody(number_of_stars) 
        stars.mass = scaled_mass

        print("centering stars")
        stars.move_to_center()
        print("scaling stars to virial equilibrium")
        stars.scale_to_standard(smoothing_length_squared
                                    = gravity.parameters.epsilon_squared)

    else:

        # Read the input data.  Units are dynamical (sorry).
        # Format:  id  mass  pos[3]  vel[3]

        print("reading file", infile)

        id = []
        mass = []
        pos = []
        vel = []

        f = open(infile, 'r')
        count = 0
        for line in f:
            if len(line) > 0:
                count += 1
                cols = line.split()
                if count == 1: snap = int(cols[0])
                elif count == 2: number_of_stars = int(cols[0])
                elif count == 3: time = float(cols[0]) | nbody_system.time
                else:
                    if len(cols) >= 8:
                        id.append(int(cols[0]))
                        mass.append(float(cols[1]))
                        pos.append((float(cols[2]),
                                    float(cols[3]), float(cols[4])))
                        vel.append((float(cols[5]),
                                    float(cols[6]), float(cols[7])))
        f.close()

        stars = datamodel.Particles(number_of_stars)
        stars.id = id
        stars.mass = mass | nbody_system.mass
        stars.position = pos | nbody_system.length
        stars.velocity = vel | nbody_system.speed
        #stars.radius = 0. | nbody_system.length

    total_mass = stars.mass.sum()
    ke = pa.kinetic_energy(stars)
    kT = ke/(1.5*number_of_stars)

    if number_of_binaries > 0:

        # Turn selected stars into binary components.
        # Only tested for equal-mass case.

        kep = Kepler(redirection = "none")
        kep.initialize_code()

        added_mass = 0.0 | nbody_system.mass

        # Work with energies rather than semimajor axes.

        Emin = 10*kT
        Emax = 20*kT
        ecc = 0.1

        nbin = 0
        companion_base_id = 100*(number_of_stars//10)

        for i in range(0, number_of_stars,
                       number_of_stars//number_of_binaries):

            # Star i is CM, becomes component, add other star at end.

            nbin += 1

            mass = stars[i].mass
            new_mass = numpy.random.uniform()*mass	# uniform q
            mbin = mass + new_mass
            fac = new_mass/mbin
            E = Emin + numpy.random.uniform()*(Emax-Emin)
            a = 0.5*nbody_system.G*mass*new_mass/E

            kep.initialize_from_elements(mbin, a, ecc)

            # Binaries should be approaching in order to be picked up
            # by multiples.
            
            kep.advance_to_apastron()
            kep.advance_to_radius(a)

            dr = quantities.AdaptingVectorQuantity()
            dr.extend(kep.get_separation_vector())
            dv = quantities.AdaptingVectorQuantity()
            dv.extend(kep.get_velocity_vector())

            newstar = datamodel.Particles(1)
            newstar.mass = new_mass
            newstar.position = stars[i].position + (1-fac)*dr
            newstar.velocity = stars[i].velocity + (1-fac)*dv

            stars[i].position = stars[i].position - fac*dr
            stars[i].velocity = stars[i].velocity - fac*dv

            newstar.id = companion_base_id + stars[i].id

            stars.add_particles(newstar)
            added_mass += new_mass

            if nbin >= number_of_binaries: break

        kep.stop()

        print('created', nbin, 'binaries')
        sys.stdout.flush()

        stars.mass = stars.mass * total_mass/(total_mass+added_mass)
        number_of_stars += nbin

    # Set dynamical radii (assuming virial equilibrium and standard
    # units).  Note that this choice should be refined, and updated
    # as the system evolves.  Probably the choice of radius should be
    # made entirely in the multiples module.  TODO.  In these units,
    # M = 1 and <v^2> = 0.5, so the mean 90-degree turnaround impact
    # parameter is
    #
    #		b_90 = G (m_1+m_2) / vrel^2
    #		     = 2 <m> / 2<v^2>
    #		     = 2 / N			for equal masses
    #
    # Taking r_i = m_i / 2<v^2> = m_i in virial equilibrium means
    # that, approximately, "contact" means a 90-degree deflection (r_1
    # + r_2 = b_90).  A more conservative choice with r_i less than
    # this value will isolate encounters better, but also place more
    # load on the large-N dynamical module.

    stars.radius = stars.mass.number | nbody_system.length

    time = 0.0 | nbody_system.time
    # print "IDs:", stars.id.number

    print("recentering stars")
    stars.move_to_center()
    sys.stdout.flush()

    #-----------------------------------------------------------------

    if softening_length < 0.0 | nbody_system.length:

        # Use ~interparticle spacing.  Assuming standard units here.  TODO

        eps2 = 0.25*(float(number_of_stars))**(-0.666667) \
			| nbody_system.length**2
    else:
        eps2 = softening_length*softening_length
    print('softening length =', eps2.sqrt())

    gravity.parameters.timestep_parameter = accuracy_parameter
    gravity.parameters.epsilon_squared = eps2
    gravity.parameters.use_gpu = use_gpu
    # gravity.parameters.manage_encounters = manage_encounters

    print('')
    print("adding particles")
    # print stars
    sys.stdout.flush()
    gravity.particles.add_particles(stars)
    gravity.commit_particles()

    print('')
    print("number_of_stars =", number_of_stars)
    sys.stdout.flush()

    # Channel to copy values from the code to the set in memory.
    channel = gravity.particles.new_channel_to(stars)

    stopping_condition = gravity.stopping_conditions.collision_detection
    stopping_condition.enable()

    # Debugging: prevent the multiples code from being called.
    if 0:
        stopping_condition.disable()
        print('stopping condition disabled')
        sys.stdout.flush()

    # -----------------------------------------------------------------
    # Create the coupled code and integrate the system to the desired
    # time, managing interactions internally.

    kep = init_kepler(stars[0], stars[1])
    multiples_code = multiples.Multiples(gravity, new_smalln, kep)

    multiples_code.neighbor_perturbation_limit = 0.1
    #multiples_code.neighbor_distance_factor = 2.0
    multiples_code.neighbor_veto = True
    multiples_code.global_debug = debug_level

    print('')
    print('multiples_code.initial_scale_factor =', \
        multiples_code.initial_scale_factor)
    print('multiples_code.neighbor_perturbation_limit =', \
        multiples_code.neighbor_perturbation_limit)
    print('multiples_code.neighbor_veto =', \
        multiples_code.neighbor_veto)
    print('multiples_code.final_scale_factor =', \
        multiples_code.final_scale_factor)
    print('multiples_code.initial_scatter_factor =', \
        multiples_code.initial_scatter_factor)
    print('multiples_code.final_scatter_factor =', \
        multiples_code.final_scatter_factor)
    print('multiples_code.retain_binary_apocenter =', \
        multiples_code.retain_binary_apocenter)
    print('multiples_code.wide_perturbation_limit =', \
        multiples_code.wide_perturbation_limit)

    # Find initial binaries.
    
    gravity.parameters.zero_step_mode = 1
    print('\nidentifying initial binaries')
    multiples_code.evolve_model(time)
    gravity.parameters.zero_step_mode = 0
    
    pre = "%%% "
    E0,cpu0 = print_log(pre, time, multiples_code)

    print("evolving to time =", end_time, \
            "in steps of", delta_t)

    while time < end_time:

        time += delta_t
        multiples_code.evolve_model(time)

        # Copy values from the module to the set in memory.

        channel.copy()
    
        # Copy the index (ID) as used in the module to the id field in
        # memory.  The index is not copied by default, as different
        # codes may have different indices for the same particle and
        # we don't want to overwrite silently.

        channel.copy_attribute("index_in_code", "id")

        print_log(pre, time, multiples_code, E0, cpu0)
        sys.stdout.flush()

    #-----------------------------------------------------------------

    if not outfile == None:

        # Write data to a file.

        f = open(outfile, 'w')

        #--------------------------------------------------
        # Need to save top-level stellar data and parameters.
        # Need to save multiple data and parameters.

        f.write('%.15g\n'%(time.number))
        for s in multiples_code.stars: write_star(s, f)

        #--------------------------------------------------

        f.close()
        print('wrote file', outfile)

    print('')
    gravity.stop()

def write_star(s, f):
    x,y,z = s.position.number
    vx,vy,vz = s.velocity.number
    f.write('%d %.15g %.15g %.15g %.15g %.15g %.15g %.15g\n' \
		%(s.id, s.mass.number, x, y, z, vx, vy, vz))

if __name__ == '__main__':

    print('\ncommand line:', end=' ')
    for a in sys.argv: print(a, end=' ')
    print('\n')

    infile = None
    outfile = None
    N = 100
    Nbin = 10
    t_end = 10.0 | nbody_system.time
    delta_t = 1.0 | nbody_system.time
    n_workers = 2
    use_gpu = 0
    gpu_worker = 0
    salpeter = 0
    accuracy_parameter = 0.1
    softening_length = 0  | nbody_system.length
    random_seed = 42
    manage_encounters = 1
    debug_level = 1

    try:
        opts, args = getopt.getopt(sys.argv[1:], "a:b:c:d:D:e:f:-F:gGn:s:St:w:")
    except getopt.GetoptError as err:
        print(str(err))
        sys.exit(1)

    for o, a in opts:
        if o == "-a":
            accuracy_parameter = float(a)
        elif o == "-b":
            Nbin = int(a)
        elif o == "-c":
            manage_encounters = int(a)
        elif o == "-d":
            delta_t = float(a) | nbody_system.time 
        elif o == "-D":
            debug_level = int(a)
        elif o == "-e":
            softening_length = float(a) | nbody_system.length
        elif o == "-f":
            infile = a
        elif o == "-F":
            outfile = a
        elif o == "-g":
            use_gpu = 0
        elif o == "-G":
            use_gpu = 0
            gpu_worker = 0
        elif o == "-n":
            N = int(a)
        elif o == "-s":
            random_seed = int(a)
        elif o == "-S":
            salpeter = 1
        elif o == "-t":
            t_end = float(a) | nbody_system.time
        elif o == "-w":
            n_workers = int(a)
        else:
            print("unexpected argument", o)

    assert is_mpd_running()
    run_ph4(infile, outfile,
            N, Nbin, t_end, delta_t, n_workers,
            use_gpu, gpu_worker,
            salpeter, accuracy_parameter, softening_length,
            manage_encounters, random_seed, debug_level)



#!/bin/env python

import collections
import getopt
import numpy
import os
import random
import sys
import unittest 
from time import process_time as cputime
from time import time as wallclocktime

from amuse.community.ph4.interface import ph4 as grav
from amuse.units import nbody_system
from amuse.units import units
from amuse.units.quantities import zero

from amuse import datamodel
from amuse.datamodel import particle_attributes as pa
from amuse.rfi.core import is_mpd_running
from amuse.ic.plummer import new_plummer_model
from amuse.ic.salpeter import new_salpeter_mass_distribution_nbody

def print_log(pre, time, gravity, E0 = 0.0 | nbody_system.energy,
              cpu0 = 0.0, wall0 = 0.0):
    cpu = cputime()
    wall = wallclocktime()
    N = len(gravity.particles)
    M = gravity.total_mass

    print('')
    print(pre+"time=", time.number)
    print(pre+"cpu=", cpu-cpu0)
    print(pre+"wall=", wall-wall0)
    print(pre+"Ntot=", N)
    print(pre+"mass=", M.number)
    E = 0 | nbody_system.energy

    if N > 0:
        U = gravity.potential_energy
        T = gravity.kinetic_energy
        Etop = T + U
        E = Etop
        if E0 == 0 | nbody_system.energy: E0 = E
        Rvir = -0.5*M*M/U
        Q = -T/U
        com = pa.center_of_mass(gravity.particles)
        comv = pa.center_of_mass_velocity(gravity.particles)
        if N > 15:
            dcen,rcore,rhocore \
              = pa.densitycentre_coreradius_coredens(gravity.particles)
        else:
            dcen = com
            rcore = zero
            rhocore = zero
            cmx,cmy,cmz = dcen
        lagr,mf = pa.LagrangianRadii(gravity.particles, cm=dcen)  # no units!

        print(pre+"Etot=", E.number)
        print(pre+"dE/E=", E/E0 - 1)
        print(pre+"Rvir=", Rvir.number)
        print(pre+"Qvir=", Q)
        cmx,cmy,cmz = com
        print(pre+"cmpos[3]= %.8f %.8f %.8f" % (cmx.number, cmy.number, cmz.number))
        cmx,cmy,cmz = comv
        print(pre+"cmvel[3]= %.8f %.8f %.8f" % (cmx.number, cmy.number, cmz.number))
        cmx,cmy,cmz = dcen
        print(pre+"dcpos[3]= %.8f %.8f %.8f" % (cmx.number, cmy.number, cmz.number))
        print(pre+"Rcore=", rcore.number)
        print(pre+"Mlagr[9]=", end=' ')
        for m in mf: print("%.4f" % (m), end=' ')
        print('')
        print(pre+"Rlagr[9]=", end=' ')
        for r in lagr.number: print("%.8f" % (r), end=' ')
        print('')

    sys.stdout.flush()
    return E,cpu,wall

def run_ph4(infile = None, number_of_stars = 40,
             end_time = 10 | nbody_system.time,
             delta_t = 1 | nbody_system.time,
             n_workers = 1, use_gpu = 1, gpu_worker = 1, gpu_id = -1,
             accuracy_parameter = 0.1,
             softening_length = -1 | nbody_system.length,
             manage_encounters = 1):

    if infile != None: print("input file =", infile)
    print("end_time =", end_time.number)
    print("delta_t =", delta_t.number)
    print("n_workers =", n_workers)
    print("use_gpu =", use_gpu)
    print("manage_encounters =", manage_encounters)
    print("\ninitializing the gravity module")
    sys.stdout.flush()

    # Note that there are actually three GPU options to test:
    #
    #	1. use the GPU code and allow GPU use (default)
    #	2. use the GPU code but disable GPU use (-g)
    #	3. use the non-GPU code (-G)

    #print "1"; sys.stdout.flush()
    
    gpu = 0
    if gpu_worker == 1:
        try:
            gravity = grav(number_of_workers = n_workers,
                           redirection = "none", mode = "gpu")
            #              debugger='valgrind')
            gpu = 1
        except Exception as ex:
            print('*** GPU worker code not found. Reverting to non-GPU code. ***')
            gpu = 0

    if gpu == 0:
        gravity = grav(number_of_workers = n_workers,
                       redirection = "none")
        #              debugger='valgrind')

    #print "2"; sys.stdout.flush()
    gravity.initialize_code()

    #print "3"; sys.stdout.flush()
    gravity.parameters.set_defaults()

    gravity.parameters.gpu_id = gpu_id

    #-----------------------------------------------------------------

    #print "4"; sys.stdout.flush()
    if infile == None:

        print("making a Plummer model")
        stars = new_plummer_model(number_of_stars)

        id = numpy.arange(number_of_stars)
        stars.id = id+1

        print("setting particle masses and radii")
        if number_of_stars > 0:
            stars.mass = (1.0 / number_of_stars) | nbody_system.mass
        if 0:
            scaled_mass = new_salpeter_mass_distribution_nbody(number_of_stars) 
            stars.mass = scaled_mass
        stars.radius = 0.0 | nbody_system.length

        print("centering stars")
        stars.move_to_center()
        if 0:
            print("scaling stars to virial equilibrium")
            stars.scale_to_standard(smoothing_length_squared
                                    = gravity.parameters.epsilon_squared)

        time = 0.0 | nbody_system.time
        sys.stdout.flush()

    else:

        # Read the input data.  Units are dynamical.

        print("reading file", infile)

        id = []
        mass = []
        pos = []
        vel = []

        f = open(infile, 'r')
        count = 0
        for line in f:
            if len(line) > 0:
                count += 1
                cols = line.split()
                if count == 1: snap = int(cols[0])
                elif count == 2: number_of_stars = int(cols[0])
                elif count == 3: time = float(cols[0]) | nbody_system.time
                else:
                    if len(cols) >= 8:
                        id.append(int(cols[0]))
                        mass.append(float(cols[1]))
                        pos.append((float(cols[2]),
                                    float(cols[3]), float(cols[4])))
                        vel.append((float(cols[5]),
                                    float(cols[6]), float(cols[7])))
        f.close()

        stars = datamodel.Particles(number_of_stars)
        stars.id = id
        stars.mass = mass | nbody_system.mass
        stars.position = pos | nbody_system.length
        stars.velocity = vel | nbody_system.speed
        stars.radius = 0. | nbody_system.length

    # print "IDs:", stars.id.number
    sys.stdout.flush()

    #-----------------------------------------------------------------

    #print "5"; sys.stdout.flush()
    if softening_length == -1 | nbody_system.length and number_of_stars > 0:
        eps2 = 0.25*(float(number_of_stars))**(-0.666667) \
                | nbody_system.length**2
    else:
        eps2 = softening_length*softening_length

    #print "6"; sys.stdout.flush()
    gravity.parameters.timestep_parameter = accuracy_parameter
    gravity.parameters.epsilon_squared = eps2
    gravity.parameters.use_gpu = use_gpu
    gravity.parameters.manage_encounters = manage_encounters

    print("adding particles")
    # print stars
    sys.stdout.flush()
    gravity.particles.add_particles(stars)
    gravity.commit_particles()

    print('Taking step')
    sys.stdout.flush()
    gravity.evolve_model(time)

    print('')
    print("number_of_stars =", number_of_stars)
    print("evolving to time =", end_time.number, \
          "in steps of", delta_t.number)
    sys.stdout.flush()

    E0,cpu0,wall0 = print_log('', time, gravity)
    
    # Channel to copy values from the code to the set in memory.
    channel = gravity.particles.new_channel_to(stars)

    stopping_condition = gravity.stopping_conditions.collision_detection
    stopping_condition.enable()
    
    while time < end_time:
        time += delta_t
        gravity.evolve_model(time)

        # Ensure that the stars list is consistent with the internal
        # data in the module.

        ls = len(stars)

        # Update the bookkeeping: synchronize stars with the module data.

        try:
            gravity.update_particle_set()
            gravity.particles.synchronize_to(stars)
        except:
            pass
    
        # Copy values from the module to the set in memory.

        channel.copy()
    
        # Copy the index (ID) as used in the module to the id field in
        # memory.  The index is not copied by default, as different
        # codes may have different indices for the same particle and
        # we don't want to overwrite silently.

        channel.copy_attribute("index_in_code", "id")

        if stopping_condition.is_set():
            star1 = stopping_condition.particles(0)[0]
            star2 = stopping_condition.particles(1)[0]
            print('\nstopping condition set at time', \
                gravity.get_time().number,'for:\n')
            print(star1)
            print('')
            print(star2)
            print('')
            raise Exception("no encounter handling")

        if len(stars) != ls:
            if 0:
                print("stars:")
                for s in stars:
                    print(" ", s.id.number, s.mass.number, \
                    s.x.number, s.y.number, s.z.number)
            else:
                print("number of stars =", len(stars))
            sys.stdout.flush()

        print_log('', time, gravity, E0, cpu0, wall0)
        sys.stdout.flush()

    print('')
    gravity.stop()

if __name__ == '__main__':

    infile = None
    N = 250
    t_end = 5.0 | nbody_system.time
    delta_t = 1.0 | nbody_system.time
    n_workers = 1
    use_gpu = 1
    gpu_worker = 1
    gpu_id = -1
    accuracy_parameter = 0.1
    softening_length = -1  | nbody_system.length
    random_seed = -1
    manage_encounters = 1

    try:
        opts, args = getopt.getopt(sys.argv[1:], "a:c:d:e:f:gGi:n:s:t:w:")
    except getopt.GetoptError as err:
        print(str(err))
        sys.exit(1)

    for o, a in opts:
        if o == "-a":
            accuracy_parameter = float(a)
        elif o == "-c":
            manage_encounters = int(a)
        elif o == "-d":
            delta_t = float(a) | nbody_system.time 
        elif o == "-e":
            softening_length = float(a) | nbody_system.length
        elif o == "-f":
            infile = a
        elif o == "-g":
            use_gpu = 0
        elif o == "-G":
            use_gpu = 0
            gpu_worker = 0
        elif o == "-i":
            gpu_id = int(a)
        elif o == "-n":
            N = int(a)
        elif o == "-s":
            random_seed = int(a)
        elif o == "-t":
            t_end = float(a) | nbody_system.time
        elif o == "-w":
            n_workers = int(a)
        else:
            print("unexpected argument", o)

    if random_seed <= 0:
        numpy.random.seed()
        random_seed = numpy.random.randint(1, pow(2,31)-1)
    numpy.random.seed(random_seed)
    print("random seed =", random_seed)


    #os.system('env')


    assert is_mpd_running()
    run_ph4(infile, N, t_end, delta_t, n_workers,
             use_gpu, gpu_worker, gpu_id,
             accuracy_parameter, softening_length,
             manage_encounters)



import collections
import getopt
import numpy
import os
import random
import sys
import unittest
import pickle

from time import clock
from time import gmtime
from time import mktime

from amuse.community.ph4.interface import ph4 as grav
from amuse.community.smalln.interface import SmallN
from amuse.community.kepler.interface import Kepler
from amuse.couple import multiples

from amuse.units import nbody_system
from amuse.units import units
from amuse.units import quantities

from amuse import datamodel
from amuse.datamodel import particle_attributes as pa
from amuse.rfi.core import is_mpd_running
from amuse.ic.plummer import new_plummer_model
from amuse.ic.salpeter import new_salpeter_mass_distribution_nbody

from amuse import io

from utils import *

def make_nbody(number_of_stars = 100, time = 0.0,
               n_workers = 1, use_gpu = 1, gpu_worker = 1,
               salpeter = 0,
               delta_t = 1.0 | nbody_system.time,
               timestep_parameter = 0.1,
               softening_length = 0.0 | nbody_system.length,
               random_seed = 1234):

    # Make an N-body system, print out some statistics on it, and save
    # it in a restart file.  The restart file name is of the form
    # 't=nnnn.n.xxx', where the default time is 0.0.

    if random_seed <= 0:
        numpy.random.seed()
        random_seed = numpy.random.randint(1, pow(2,31)-1)
    numpy.random.seed(random_seed)
    print("random seed =", random_seed)

    init_smalln()

    # Note that there are actually three GPU options:
    #
    #	1. use the GPU code and allow GPU use (default)
    #	2. use the GPU code but disable GPU use (-g)
    #	3. use the non-GPU code (-G)

    if gpu_worker == 1:
        try:
            gravity = grav(number_of_workers = n_workers,
                           redirection = "none", mode = "gpu")
        except Exception as ex:
            gravity = grav(number_of_workers = n_workers,
                           redirection = "none")
    else:
        gravity = grav(number_of_workers = n_workers,
                       redirection = "none")

    gravity.initialize_code()
    gravity.parameters.set_defaults()

    #-----------------------------------------------------------------

    # Make a standard N-body system.

    print("making a Plummer model")
    stars = new_plummer_model(number_of_stars)

    id = numpy.arange(number_of_stars)
    stars.id = id+1

    print("setting particle masses and radii")
    if salpeter == 0:
        print('equal masses')
        total_mass = 1.0 | nbody_system.mass
        scaled_mass = total_mass / number_of_stars
    else:
        print('salpeter mass function')
        mmin = 0.5 | nbody_system.mass
        mmax = 10.0 | nbody_system.mass
        scaled_mass = new_salpeter_mass_distribution_nbody(number_of_stars,
                                                           mass_min = mmin,
                                                           mass_max = mmax)
        
    stars.mass = scaled_mass

    print("centering stars")
    stars.move_to_center()
    print("scaling stars to virial equilibrium")
    stars.scale_to_standard(smoothing_length_squared
                            = gravity.parameters.epsilon_squared)

    # Set dynamical radii (assuming virial equilibrium and standard
    # units).  Note that this choice should be refined, and updated
    # as the system evolves.  Probably the choice of radius should be
    # made entirely in the multiples module.  TODO.  In these units,
    # M = 1 and <v^2> = 0.5, so the mean 90-degree turnaround impact
    # parameter is
    #
    #		b_90 = G (m_1+m_2) / vrel^2
    #		     = 2 <m> / 2<v^2>
    #		     = 2 / N			for equal masses
    #
    # Taking r_i = m_i / 2<v^2> = m_i in virial equilibrium means
    # that, approximately, "contact" means a 90-degree deflection (r_1
    # + r_2 = b_90).  A more conservative choice with r_i less than
    # this value will isolate encounters better, but also place more
    # load on the large-N dynamical module.

    stars.radius = 0.5*stars.mass.number | nbody_system.length

    time = 0.0 | nbody_system.time
    # print "IDs:", stars.id.number

    print("recentering stars")
    stars.move_to_center()
    sys.stdout.flush()

    #-----------------------------------------------------------------

    if softening_length < 0.0 | nbody_system.length:

        # Use ~interparticle spacing.  Assuming standard units here.  TODO

        softening_length = 0.5*float(number_of_stars)**(-0.3333333) \
				| nbody_system.length
    print('softening length =', softening_length)

    gravity.parameters.timestep_parameter = timestep_parameter
    gravity.parameters.epsilon_squared = softening_length*softening_length
    gravity.parameters.use_gpu = use_gpu

    print('')
    print("adding particles")
    # print stars
    sys.stdout.flush()
    gravity.particles.add_particles(stars)
    gravity.commit_particles()

    print('')
    print("number_of_stars =", number_of_stars)
    sys.stdout.flush()

    # Channel to copy values from the code to the set in memory.
    channel = gravity.particles.new_channel_to(stars)

    stopping_condition = gravity.stopping_conditions.collision_detection
    stopping_condition.enable()

    # -----------------------------------------------------------------
    # Create the coupled code and integrate the system to the desired
    # time, managing interactions internally.

    kep = init_kepler(stars[0], stars[1])
    multiples_code = multiples.Multiples(gravity, new_smalln, kep)

    multiples_code.neighbor_perturbation_limit = 0.1
    multiples_code.neighbor_veto = True

    print('')
    print('multiples_code.initial_scale_factor =', \
        multiples_code.initial_scale_factor)
    print('multiples_code.neighbor_perturbation_limit =', \
        multiples_code.neighbor_perturbation_limit)
    print('multiples_code.neighbor_veto =', \
        multiples_code.neighbor_veto)
    print('multiples_code.final_scale_factor =', \
        multiples_code.final_scale_factor)
    print('multiples_code.initial_scatter_factor =', \
        multiples_code.initial_scatter_factor)
    print('multiples_code.final_scatter_factor =', \
        multiples_code.final_scatter_factor)
    print('multiples_code.retain_binary_apocenter =', \
        multiples_code.retain_binary_apocenter)
    print('multiples_code.wide_perturbation_limit =', \
        multiples_code.wide_perturbation_limit)

    # Take a dummy step, just in case...

    multiples_code.evolve_model(time)

    # Copy values from the module to the set in memory.

    channel.copy()
    
    # Copy the index (ID) as used in the module to the id field in
    # memory.  The index is not copied by default, as different
    # codes may have different indices for the same particle and
    # we don't want to overwrite silently.

    channel.copy_attribute("index_in_code", "id")

    pre = "%%% "
    E0,cpu0 = print_log(pre, time, multiples_code)
    sys.stdout.flush()

    # file = 't='+'{:07.2f}'.format(time.number)        # fails in Python 2.6
    file = 't=%07.2f'%time.number
    write_state_to_file(time, stars, gravity, multiples_code,
                        file, delta_t, E0, cpu0)
    tree_copy = multiples_code.root_to_tree.copy()

    del multiples_code
    sys.stdout.flush()

    gravity.stop()
    kep.stop()
    stop_smalln()
    print('')


if __name__ == '__main__':

    # Defaults:

    N = 1000
    time = 0.0 | nbody_system.time
    delta_t = 1.0 | nbody_system.time
    n_workers = 1
    use_gpu = 1
    gpu_worker = 1
    salpeter = 0
    timestep_parameter = 0.1
    softening_length = 0.0 | nbody_system.length
    random_seed = -1

    try:
        opts, args = getopt.getopt(sys.argv[1:], "n:st:")
    except getopt.GetoptError as err:
        print(str(err))
        sys.exit(1)

    for o, a in opts:
        if o == "-n":
            N = int(a)
        elif o == "-s":
            salpeter = 1
        elif o == "-t":
            time = float(a) | nbody_system.time
        else:
            print("unexpected argument", o)

    assert is_mpd_running()

    make_nbody(N, time, n_workers, use_gpu, gpu_worker,
               salpeter, delta_t, timestep_parameter, softening_length,
               random_seed)



import collections
import getopt
import numpy
import os
import random
import sys
import pickle
import math

import unittest
from time import clock

from amuse.community.ph4.interface import ph4 as grav
from amuse.community.smalln.interface import SmallN
from amuse.community.kepler.interface import Kepler
from amuse.couple import multiples

from amuse.units import nbody_system
from amuse.units import units
from amuse.units import quantities

from amuse import datamodel
from amuse.datamodel import particle_attributes as pa
from amuse.rfi.core import is_mpd_running
from amuse.ic.plummer import new_plummer_model
from amuse.ic.salpeter import new_salpeter_mass_distribution_nbody

from amuse import io
from utils import *

def handle_callback (time, star1, star2):
    print('')
    print('    callback')
    print('   ', time)
    print('   ', star1)
    print('   ', star2)
    print('')
    return True		# NOTE: returning False will skip this encounter

def run_ph4(initial_file = None,
            end_time = 0 | nbody_system.time,
            input_delta_t = 0.0 | nbody_system.time,
            input_Delta_t = 1.0 | nbody_system.time,
            input_timestep_parameter = 0.0,
            input_softening_length = -1.0 | nbody_system.length,
            n_workers = 1, use_gpu = 1, gpu_worker = 1,
            use_multiples = True,
            save_restart = False,
            strict_restart = False
        ):

    # Read an N-body system from a file and run it to the specified
    # time using the specified steps.  Print log information and
    # optionally save a restart file after every step.  If the
    # specified time is less than the time in the initial file, don't
    # take a step, but still print out the log info.  (Hence run_ph4
    # also functions like Starlab sys_stats.)

    print("initial_file =", initial_file)
    print("end_time =", end_time.number)
    print("n_workers =", n_workers)
    print("use_gpu =", use_gpu)
    print("use_multiples =", use_multiples)
    print("save_restart =", save_restart)
    print("strict_restart =", strict_restart)
    print("\ninitializing the gravity module")
    sys.stdout.flush()

    init_smalln()
    
    # Note that there are actually three GPU options:
    #
    #	1. use the GPU code and allow GPU use (default)
    #	2. use the GPU code but disable GPU use (-g)
    #	3. use the non-GPU code (-G)

    if gpu_worker == 1:
        try:
            gravity = grav(number_of_workers = n_workers,
                           redirection = "none", mode = "gpu")
        except Exception as ex:
            gravity = grav(number_of_workers = n_workers,
                           redirection = "none")
    else:
        gravity = grav(number_of_workers = n_workers,
                       redirection = "none")

    gravity.initialize_code()
    gravity.parameters.set_defaults()

    kep = Kepler(None, redirection = "none")
    kep.initialize_code()

    stars, time, delta_t, E0, cpu0, multiples_code \
        	= read_state_from_file(initial_file, gravity, kep)

    # Allow overrides of the restored data (OK for delta_t, NOT
    # recommended for timestep_parameter or softening_length).  Note
    # that reading the state also commits the particles, and hence
    # calculates the initial time steps.  Probably should reinitialize
    # if timestep_parameter or softening_length are changed.  TODO

    if input_delta_t.number > 0:
        if input_delta_t != delta_t:
            print('modifying delta_t from stored', delta_t, \
		  'to input', input_delta_t)
            delta_t = input_delta_t
    else:
        print("using stored delta_t =", delta_t)

    print(input_timestep_parameter)
    print(gravity.parameters.timestep_parameter)

    if input_timestep_parameter > 0:
        if input_timestep_parameter != gravity.parameters.timestep_parameter:
            print('modifying timestep_parameter from stored', \
            	  gravity.parameters.timestep_parameter, \
		  'to input', input_timestep_parameter)
            gravity.parameters.timestep_parameter \
		= input_timestep_parameter
    else:
        print('timestep_parameter =', gravity.parameters.timestep_parameter)

    if input_softening_length.number >= 0:
        if input_softening_length*input_softening_length \
		!= gravity.parameters.epsilon_squared:
            print('modifying softening_length from stored', \
            	  gravity.parameters.epsilon_squared.sqrt(), \
		  'to input', input_softening_length)
            gravity.parameters.epsilon_squared \
                = softening_length*softening_length
    else:
        print('softening length =', gravity.parameters.epsilon_squared.sqrt())

    gravity.parameters.use_gpu = use_gpu
    gravity.parameters.begin_time = time

    if 0:
        print('')
        print(gravity.parameters.begin_time)
        print(stars.mass)
        #print stars.position
        for s in stars:
            print('%.18e %.18e %.18e' % (s.x.number, s.y.number, s.z.number))
        print(stars.velocity)

    channel = gravity.particles.new_channel_to(stars)

    if use_multiples:
        stopping_condition = gravity.stopping_conditions.collision_detection
        stopping_condition.enable()

    gravity.parameters.force_sync = 1	# end exactly at the specified time

    pre = "%%% "
    print_log(pre, time, multiples_code, E0, cpu0)

    tsave = time + Delta_t
    save_file = ''

    while time < end_time:

        time += delta_t
        multiples_code.evolve_model(time) #, callback=handle_callback)

        # Copy values from the module to the set in memory.

        channel.copy()
    
        # Copy the index (ID) as used in the module to the id field in
        # memory.  The index is not copied by default, as different
        # codes may have different indices for the same particle and
        # we don't want to overwrite silently.

        channel.copy_attribute("index_in_code", "id")

        # Write log information.

        print_log(pre, time, multiples_code, E0, cpu0)
        sys.stdout.flush()

        # Optionally create a restart file.

        if save_restart and time >= tsave:
            #save_file = 't='+'{:07.2f}'.format(time.number) # not in Python 2.6
            save_file = 't=%07.2f'%time.number
            write_state_to_file(time, stars, gravity, multiples_code,
                                save_file, delta_t, E0, cpu0)
            sys.stdout.flush()
            tsave += Delta_t
            if strict_restart: break

    gravity.stop()
    kep.stop()
    stop_smalln()

    return time, save_file


def print_help():
    print("Options:")
    print("    -d set log output interval [1.0]")
    print("    -h print this help message")
    print("    -i set initial file name [t=0000.0]")
    print("    -m suppress multiples  [False]")
    print("    -s save restart files [False]")
    print("    -t set final time [0.0]")

if __name__ == '__main__':

    initial_file = 't=0000.0'
    t_end = 0.0 | nbody_system.time	# default is to print log and exit
    delta_t = 0.0 | nbody_system.time	# log output time scale
    Delta_t = 1.0 | nbody_system.time	# restart output time scale
    Delta_t_set = False
    timestep_parameter = -1
    softening_length = -1  | nbody_system.length
    n_workers = 1
    use_gpu = 1
    gpu_worker = 1
    use_multiples = True
    save_restart = False
    strict_restart = False

    try:
        opts, args = getopt.getopt(sys.argv[1:], "d:D:hi:msSt:")
    except getopt.GetoptError as err:
        print(str(err))
        sys.exit(1)

    for o, a in opts:
        if o == "-d":
            delta_t = float(a) | nbody_system.time
        elif o == "-D":
            Delta_t = float(a) | nbody_system.time
            Delta_t_set = True
        elif o == "-h":
            print_help()
            sys.exit(1)
        elif o == "-i":
            initial_file = a
        elif o == "-m":
            use_multiples = False
        elif o == "-s":
            save_restart = True
        elif o == "-S":
            strict_restart = True
            save_restart = True
        elif o == "-t":
            t_end = float(a) | nbody_system.time
        else:
            print("unexpected argument", o)
            print_help()
            sys.exit(1)

    if not save_restart: strict_restart = False
    if strict_restart: save_restart = True
    if not Delta_t_set: Delta_t = delta_t

    # Code implicitly assumes that delta_t and Delta_t are
    # commensurate. Check that here.

    if math.fmod(Delta_t.number, delta_t.number) != 0:
        x = math.floor(Delta_t/delta_t)
        Delta_t = x*delta_t
        print('reset Delta_t to', Delta_t)

    assert is_mpd_running()

    # In non-strict_mode, OK to let run_ph4 loop over steps. If
    # strict_restart is True, handle the loop here and force a new
    # restart at every step -- seems substantially slower, but
    # should be reproducible.

    if (not strict_restart):
        run_ph4(initial_file, t_end, delta_t, Delta_t,
                timestep_parameter, softening_length,
                n_workers, use_gpu, gpu_worker,
                use_multiples,
                save_restart, strict_restart)
    else:
        t = -1.0 | nbody_system.time
        while t < t_end:
            t, initial_file = run_ph4(initial_file, t_end, delta_t, Delta_t,
                                      timestep_parameter, softening_length,
                                      n_workers, use_gpu, gpu_worker,
                                      use_multiples,
                                      save_restart, strict_restart)
            delta_t = 0.0 | nbody_system.time
            timestep_parameter = -1
            softening_length = -1  | nbody_system.length
            print('')



import collections
import getopt
import numpy
import os
import random
import sys
import pickle
import numpy as np

import unittest
from time import clock

from amuse.community.ph4.interface import ph4 as grav
from amuse.community.smalln.interface import SmallN
from amuse.community.kepler.interface import Kepler
from amuse.couple import multiples

from amuse.units import nbody_system
from amuse.units import units
from amuse.units import quantities

from amuse import datamodel
from amuse.datamodel import particle_attributes as pa
from amuse.rfi.core import is_mpd_running
from amuse.ic.plummer import new_plummer_model
from amuse.ic.salpeter import new_salpeter_mass_distribution_nbody

from amuse import io

SMALLN = None
def new_smalln():
    SMALLN.reset()
    return SMALLN
    
def init_smalln():
    global SMALLN
    sys.stdout.flush()
    SMALLN = SmallN()#redirection='none')
    sys.stdout.flush()
    SMALLN.parameters.timestep_parameter = 0.1
    #SMALLN.parameters.cm_index = 2001		# don't set this here!!
    sys.stdout.flush()

def stop_smalln():
    global SMALLN
    SMALLN.stop()

def init_kepler(star1, star2):        
    try:
        star1.mass.value_in(units.kg) # see if SI units, throw exception if not
        unit_converter \
            = nbody_system.nbody_to_si(star1.mass + star2.mass,
                                       (star2.position-star1.position).length())
    except Exception as ex:
        unit_converter = None
        
    kep = Kepler(unit_converter) # , redirection = "none")
    kep.initialize_code()

    return kep

def print_binaries(pre, stars, kT, limit):	# print all bound pairs with
						# binding energy >= limit*kT
    Emax = -(limit*kT).number
    m = np.array(stars.mass.number)
    pos = np.array(stars.position.number)
    vel = np.array(stars.velocity.number)
    n = len(stars)
    dx = np.zeros((n,3))
    dv = np.zeros((n,3))
    tiny = 1.e-12

    Ntop = 0
    Etop = 0.0
    print('bound top-level pairs:')
    for i in range(n):
        dx[:,:] = pos[:,:] - pos[i,:]
        dv[:,:] = vel[:,:] - vel[i,:]
        dr2 = (dx**2).sum(axis=1) + tiny
        dr2i = 1./dr2
        dr2i[i] = 0.0
        mi = m[i]
        dpot = -(m+mi)*np.sqrt(dr2i)
        dpot[:i] = 0.0
        dkin = 0.5*(dv**2).sum(axis=1)
        mu = mi*m/(mi+m)
        de = mu*(dpot + dkin)
        w = np.where(de < Emax)
        jw = np.arange(n)[w]
        if len(jw) > 0:
            for j in jw:
                print('   ', stars.id[i], stars.id[j], de[j]/kT.number, 'kT')
                Ntop += 1
                Etop += de[j]

    if Ntop == 0:
        print('(none)')

    print('')
    print(pre, 'Ntop =', Ntop)
    print(pre, 'Etop =', Etop)
    print(pre, 'Etop/kT =', Etop/kT.number)

def print_log(pre, time, gravity, E0 = 0.0 | nbody_system.energy, cpu0 = 0.0):
    cpu = clock()
    N = len(gravity.particles)
    M = gravity.total_mass
    U = gravity.potential_energy
    T = gravity.kinetic_energy
    Etop = T + U
    Nmul, Nbin, Emul = gravity.get_total_multiple_energy()
    tmp1,tmp2,Emul2 = gravity.get_total_multiple_energy2()
    Etot = Etop + Emul
    Eext = gravity.multiples_external_tidal_correction
    Eint = gravity.multiples_internal_tidal_correction
    Eerr = gravity.multiples_integration_energy_error
    Edel = Eext + Eint + Eerr
    Ecor = Etot - Edel
    if E0 == 0 | nbody_system.energy: E0 = Ecor
    Rvir = -0.5*M*M/U
    Q = -T/U
    com = pa.center_of_mass(gravity.particles)
    comv = pa.center_of_mass_velocity(gravity.particles)

    # Hop complains if we have too few particles.

    dcen = com
    if N >= 100:
        dcen,rcore,rhocore \
            = pa.densitycentre_coreradius_coredens(gravity.particles)
        cmx,cmy,cmz = dcen
        lagr,mf = pa.LagrangianRadii(gravity.particles, cm=dcen)  # no units!

    print('')
    print(pre+"time=", time.number)
    print(pre+"CPU=", cpu - cpu0)
    print(pre+"Ntot=", N)
    print(pre+"mass=", M.number)
    print(pre+"Etot=", Etot.number)
    print(pre+"Etop=", Etop.number)
    print(pre+"Eext=", Eext.number)
    print(pre+"Eint=", Eint.number)
    print(pre+"Eerr=", Eerr.number)
    print(pre+"Edel=", Edel.number)
    print(pre+"Ecor=", Ecor.number)
    print(pre+"dE=", Ecor.number - E0.number)
    print(pre+"dE/E=", Ecor/E0 - 1)
    print(pre+"Rvir=", Rvir.number)
    print(pre+"Qvir=", Q)
    cmx,cmy,cmz = com
    print(pre+"cmpos[3]= %.8f %.8f %.8f" % (cmx.number, cmy.number, cmz.number))
    cmx,cmy,cmz = comv
    print(pre+"cmvel[3]= %.8f %.8f %.8f" % (cmx.number, cmy.number, cmz.number))

    if N >= 100:
        cmx,cmy,cmz = dcen
        print(pre+"dcpos[3]= %.8f %.8f %.8f" \
            		% (cmx.number, cmy.number, cmz.number))
        print(pre+"Rcore=", rcore.number)
        print(pre+"Mcore=", (rhocore*rcore**3).number)	# fake...
        print(pre+"Mlagr[9]=", end=' ')
        for m in mf: print("%.4f" % (m), end=' ')
        print('')
        print(pre+"Rlagr[9]=", end=' ')
        for r in lagr.number: print("%.8f" % (r), end=' ')
        print('')

    kT = 2*T/(3.*N)	# 3/2 N kT = total KE, by definition
    kT0 = -2*E0/(3.*N)
    print(pre+"kT= %.8f" % (kT.number))
    Nmul,Nbin,Emul = gravity.print_multiples2(pre, kT, dcen)
    print(pre+"Nmul=", Nmul)
    print(pre+"Nbin=", Nbin)
    print(pre+"Emul= %.5f" % (Emul.number))
    print(pre+"Emul2= %.5f" % (Emul2.number))
    print(pre+"Emul/kT= %.5f" % (Emul/kT))
    print(pre+"Emul/kT0= %.5f" % (Emul/kT0))
    print(pre+"Emul/E= %.5f" % (Emul/Etot))
    print('')

    stars = gravity.particles.copy()
    schannel = gravity.particles.new_channel_to(stars)
    schannel.copy_attribute("index_in_code", "id")
    print_binaries(pre, stars, kT, 0.1)

    sys.stdout.flush()
    return Ecor,cpu


# Save code from Brewer...

def write_state_to_file(time, stars_python, gravity_code, multiples_code,
                        write_file=None,
                        delta_t=1.0|nbody_system.time,
                        E0_log=0.0, cpu0_log=0.0):

    # Function for saving everything.  If you need to track more than
    # is normally saved, should be safe to tack it onto params.

    if write_file is not None:

        particles = gravity_code.particles.copy()
        write_channel = gravity_code.particles.new_channel_to(particles)
        write_channel.copy_attribute("index_in_code", "id")

        params = {'timestep_parameter': \
                      gravity_code.parameters.timestep_parameter,
                  'epsilon_squared': gravity_code.parameters.epsilon_squared,
                  'neighbor_veto': multiples_code.neighbor_veto,
                  'multiples_external_tidal_correction': \
                      multiples_code.multiples_external_tidal_correction,
                  'multiples_integration_energy_error': \
                      multiples_code.multiples_integration_energy_error,
                  'multiples_internal_tidal_correction': \
                      multiples_code.multiples_internal_tidal_correction,
                  'model_time': multiples_code.model_time,
                  'delta_t': delta_t,
                  'EZero': E0_log,
                  'CPUZero': cpu0_log,
                  'root_index': multiples.root_index
        }
        
        for root, tree in multiples_code.root_to_tree.items():
            root_in_particles = root.as_particle_in_set(particles)
            subset = tree.get_tree_subset().copy()
            if root_in_particles is not None:
                root_in_particles.components = subset

        io.write_set_to_file(particles,
                             write_file+".stars.hdf5",
                             'amuse',
                             version='2.0',
                             append_to_file=False,
                             copy_history=False,
                             close_file=True)

        io.write_set_to_file(stars_python,
                             write_file+".stars_python.hdf5",
                             'amuse',
                             version='2.0',
                             append_to_file=False,
                             copy_history=False,
                             close_file=True)

        config = {'time': time,
                  'py_seed': pickle.dumps(random.getstate()),
                  'numpy_seed': pickle.dumps(numpy.random.get_state()),
              }
        with open(write_file + ".conf", "wb") as f:
            pickle.dump(config, f)
            f.close()
            with open(write_file + ".params", "wb") as f:
                pickle.dump(params, f)
                f.close()

        print("\nwrote state to file ", write_file, 'at time', time)


# Restore code from Brewer.

def read_state_from_file(restart_file, gravity_code, kep, MT=0):

    # Function to load from file.  If you change params in
    # write_state_to_file, make sure you match the changes here.

    stars = io.read_set_from_file(restart_file+".stars.hdf5",
                                  'amuse', version='2.0',
                                  close_file=True).copy()
    stars_python = io.read_set_from_file(
        		restart_file+".stars_python.hdf5",
        		'amuse',version='2.0',
			close_file=True).copy()
    with open(restart_file + ".params", "rb") as f:
        params = pickle.load(f)
        f.close()
    
    if MT == 0:
        root_to_tree = {}
        for root in stars:
            if hasattr(root, 'components') and not root.components is None:
                root_to_tree[root] \
                    = datamodel.trees.BinaryTreeOnParticle(root.components[0])
    else:
        root_to_tree = MT

    gravity_code.parameters.timestep_parameter = params['timestep_parameter']
    gravity_code.parameters.epsilon_squared = params['epsilon_squared']

    gravity_code.particles.add_particles(stars)
    gravity_code.commit_particles()		# sets time steps

    multiples_code = multiples.Multiples(gravity_code, new_smalln, kep)
    multiples_code.neighbor_veto = params['neighbor_veto']
    multiples_code.multiples_external_tidal_correction \
	= params['multiples_external_tidal_correction']
    multiples_code.multiples_integration_energy_error \
	= params['multiples_integration_energy_error']
    multiples_code.multiples_internal_tidal_correction \
	= params['multiples_internal_tidal_correction']
    multiples.root_index = params['root_index']
    multiples_code.root_to_tree = root_to_tree

    print("\nread state from file ", restart_file, \
          'at time', params['model_time'])

    return stars_python, params['model_time'], params['delta_t'], \
		params['EZero'], params['CPUZero'], multiples_code



#!/usr/bin/env python

import sys
from math import log10, sqrt
import numpy as np
import matplotlib.pyplot as plt
import string
import re

# Display a time sequence from a set of (not yet standard) AMUSE log
# files.  Files, colors, and styles should be separated by spaces.

if __name__ == '__main__':

    # Command-line arguments and defaults:

    infiles = []				# -f [list]
    log = 0					# -L (sets log = 1)
    outfile = ''				# -o filename or .
    quantities = []				# -q [list]
    layout = ''					# -l (rows)x(cols)
    space = 0.3					# -S space

    # Separate the point colors and the line styles.  Replicate the
    # last color/style in case of multiple files.

    colors = ['b','g','r','c','m','k']		# -c [list]
    ls = ['-']					# -s [list]

    # Parse the argument list the old-fashioned way to enable multiple
    # arguments (unaccountably not supported by getopt):

    i = 1
    n = len(sys.argv)
    while i < n:
        if sys.argv[i][0] == '-':

            if len(sys.argv[i]) > 1:

                if sys.argv[i][1] == 'c':

                    colors = []

                    # Accumulate colors until we find a leading '-' or
                    # reach the end of the list.

                    while i < n-1 and sys.argv[i+1][0] != '-':
                        i += 1
                        colors.append(sys.argv[i])

                elif sys.argv[i][1] == 'f':

                    # Accumulate names until we find a leading '-' or
                    # reach the end of the list.

                    while i < n-1 and sys.argv[i+1][0] != '-':
                        i += 1
                        infiles.append(sys.argv[i])

                elif sys.argv[i][1] == 'L':

                    log = 1

                elif sys.argv[i][1] == 'l':

                    if i < n-1 and sys.argv[i+1][0] != '-':
                        i += 1
                        layout = sys.argv[i]

                elif sys.argv[i][1] == 'o':

                    if i < n-1 and sys.argv[i+1][0] != '-':
                        i += 1
                        outfile = sys.argv[i]

                elif sys.argv[i][1] == 'q':

                    # Accumulate quantities until we find a leading
                    # '-' or reach the end of the list.

                    while i < n-1 and sys.argv[i+1][0] != '-':
                        i += 1
                        quantities.append(sys.argv[i])

                elif sys.argv[i][1] == 'S':

                    if i < n-1 and sys.argv[i+1][0] != '-':
                        i += 1
                        space = float(sys.argv[i])

                elif sys.argv[i][1] == 's':

                    ls = []

                    # Accumulate line styles until we find a leading
                    # '-' or reach the end of the list.  Problem: '-'
                    # is a legal line style, so accept a lone '-' in
                    # this case.

                    while i < n-1 \
                            and (sys.argv[i+1][0] != '-'
                                  or len(sys.argv[i+1]) == 1):
                        i += 1
                        ls.append(sys.argv[i])

                else:
                    print('unknown option', sys.argv[i])

        i += 1
    
    li = len(infiles)
    lq = len(quantities)
    if li == 0 or lq == 0:
        print('plotq -f file-list -q quantity-list')
        sys.exit(0)

    # Extend the color list if neessary.

    lc = len(colors)
    if lc < li:
        for i in range(li-lc):
            colors.append(colors[lc-1])

    # Extend the line style list if necessary.

    lt = len(ls)
    if lt < li:
        for i in range(li-lt):
            ls.append(ls[lt-1])

    plt.figure(figsize=(12,8))
    plt.subplots_adjust(wspace=space, hspace=space)

    # Determine the plot layout (mxn).

    if layout == '':

        # Automatic.

        m = int(sqrt(float(lq)))
        if m*m < lq: m += 1
        n = lq/m
        if m*n < lq: n += 1

    else:

        # User-specified:

        m,n = layout.split('x')
        m = int(m)
        n = int(n)

    i = 100*m+10*n				# assume m, n < 10
    figs = []
    for ii in range(lq):
        figs.append(plt.subplot(i+ii+1))

    print('infiles:    ', end=' ')
    for f in infiles: print(f, end=' ')
    print('')
    print('quantities: ', end=' ')
    for q in quantities: print(q, end=' ')
    print('')
    print('colors:     ', end=' ')
    for ic in range(li): print(colors[ic], end=' ')
    print('')
    print('styles:     ', end=' ')
    for i in range(li): print(ls[i], end=' ')
    print('')
    print(m, 'x', n, 'plot layout')

    ifile = 0
    for infile in infiles:			# loop over input files

        # Extract data from infile.

        tlist = []
        qlist = [[] for iq in range(lq)]	# list of lq empty lists
        nq = [1 for iq in range(lq)]		# default qlist length

        f = open(infile, 'r')
        for line in f:				# process data line by line
            if len(line) > 0:
                cols = line.split()
                if len(cols) > 0 and cols[0] == '%%%':

                    if cols[1] == 'time=':
                        tlist.append(float(cols[2]))

                    else:

                        foundq = False
                        for iq in range(lq):
                            if cols[1] == quantities[iq]+'=':
                                foundq = True
                                q = float(cols[2])
                                if log: q = log10(q)
                                qlist[iq].append(q)

                        if not foundq:
                            for iq in range(lq):
                                qq = quantities[iq]
                                if re.search(qq+'\[.*\]'+'=', cols[1]):
                                    s = re.search(qq+'\[.*\]'+'=',
                                                   cols[1]).group()
                                    # Messy! There must surely be a
                                    # better way to do this...
                                    i1 = string.index(s,'[')
                                    i2 = string.index(s,']')
                                    nq[iq] = int(s[i1+1:i2])
                                    clist = []
                                    for i in range(2,2+nq[iq]):
                                        q = float(cols[i])
                                        if log: q = log10(q)
                                        clist.append(q)
                                    qlist[iq].append(clist)

        f.close()

        # Plot the data.

        for iq in range(lq):
            qq = quantities[iq]
            if len(qlist[iq]) > 0:
                if log: qq = 'log10 '+qq
                if nq[iq] == 1:
                    figs[iq].plot(tlist, qlist[iq], colors[ifile]+ls[ifile])
                else:
                    qplot = np.array(qlist[iq])
                    for i in range(nq[iq]):
                        figs[iq].plot(tlist, qplot[:,i],
                             colors[ifile]+ls[ifile], linewidth=1.0)
                figs[iq].set_xlabel('time')
                figs[iq].set_ylabel(qq)
            else:
                print('No', qq, 'data to plot in '+infile+'.')

        ifile += 1

    # Save to a file if specified; otherwise, display to the screen.

    if outfile == '':
        plt.show()
    else:
        if outfile == '.':
            qq = ''
            for iq in range(lq):
                if iq > 0: qq += '+'
                qq += quantities[iq]
            outfile = qq+'.pdf'
        outfile = outfile.replace('/','_')
        print('saving to', outfile)
        plt.savefig(outfile)



import collections
import getopt
import numpy
import os
import random
import sys
import unittest

from amuse.community.ph4.interface import ph4 as grav
from amuse.community.smalln.interface import SmallN
from amuse.community.kepler.interface import Kepler
from amuse.couple import multiples

from amuse.units import nbody_system
from amuse.units import units

from amuse import datamodel
from amuse.datamodel import particle_attributes
from amuse.rfi.core import is_mpd_running
from amuse.ic.plummer import new_plummer_model
from amuse.ic.salpeter import new_salpeter_mass_distribution_nbody

def print_log(time, gravity, E0 = 0.0 | nbody_system.energy):
    M = gravity.total_mass
    U = gravity.potential_energy
    T = gravity.kinetic_energy
    Ebin = gravity.get_binary_energy()
    Etop = T + U
    E = Etop + Ebin
    if E0 == 0 | nbody_system.energy: E0 = E
    Rv = -0.5*M*M/U
    Q = -T/U
    print("")
    print("time =", time.number, " energy = ", E.number, \
	" dE/E0 = ", (E/E0 - 1))
    print('%s %.4f %.6f %.6f %.6f %.6f %.6f %.6f %.6f' % \
	("%%", time.number, M.number, T.number, U.number, \
         E.number, Ebin.number, Rv.number, Q))
    sys.stdout.flush()
    return E

def new_smalln():
    result = SmallN()
    result.parameters.timestep_parameter = 0.1
    #result.parameters.cm_index = 2001
    return result
        
def run_ph4(infile = None, number_of_stars = 40,
             end_time = 10 | nbody_system.time,
             delta_t = 1 | nbody_system.time,
             n_workers = 1, use_gpu = 1, gpu_worker = 1,
             accuracy_parameter = 0.1,
             softening_length = -1 | nbody_system.length,
             manage_encounters = 1, random_seed = 1234):

    if random_seed <= 0:
        numpy.random.seed()
        random_seed = numpy.random.randint(1, pow(2,31)-1)
    numpy.random.seed(random_seed)
    print("random seed =", random_seed)

    if infile != None: print("input file =", infile)
    print("end_time =", end_time.number)
    print("delta_t =", delta_t.number)
    print("n_workers =", n_workers)
    print("use_gpu =", use_gpu)
    print("manage_encounters =", manage_encounters)
    print("\ninitializing the gravity module")
    sys.stdout.flush()

    # Note that there are actually three GPU options to test:
    #
    #	1. use the GPU code and allow GPU use (default)
    #	2. use the GPU code but disable GPU use (-g)
    #	3. use the non-GPU code (-G)

    if gpu_worker == 1:
        try:
            gravity = grav(number_of_workers = n_workers,
                           redirection = "none", mode = "gpu")
        except Exception as ex:
            gravity = grav(number_of_workers = n_workers, redirection = "none")
    else:
        gravity = grav(number_of_workers = n_workers, redirection = "none")

    gravity.initialize_code()
    gravity.parameters.set_defaults()

    #-----------------------------------------------------------------

    if infile == None:

        print("making a Plummer model")
        stars = new_plummer_model(number_of_stars)

        id = numpy.arange(number_of_stars)
        stars.id = id+1

        print("setting particle masses and radii")
	#stars.mass = (1.0 / number_of_stars) | nbody_system.mass
        scaled_mass = new_salpeter_mass_distribution_nbody(number_of_stars) 
        stars.mass = scaled_mass
        stars.radius = 0.02 | nbody_system.length

        print("centering stars")
        stars.move_to_center()
        print("scaling stars to virial equilibrium")
        stars.scale_to_standard(smoothing_length_squared
                                    = gravity.parameters.epsilon_squared)

        time = 0.0 | nbody_system.time
        sys.stdout.flush()

    else:

        # Read the input data.  Units are dynamical.

        print("reading file", infile)

        id = []
        mass = []
        pos = []
        vel = []

        f = open(infile, 'r')
        count = 0
        for line in f:
            if len(line) > 0:
                count += 1
                cols = line.split()
                if count == 1: snap = int(cols[0])
                elif count == 2: number_of_stars = int(cols[0])
                elif count == 3: time = float(cols[0]) | nbody_system.time
                else:
                    if len(cols) >= 8:
                        id.append(int(cols[0]))
                        mass.append(float(cols[1]))
                        pos.append((float(cols[2]),
                                    float(cols[3]), float(cols[4])))
                        vel.append((float(cols[5]),
                                    float(cols[6]), float(cols[7])))
        f.close()

        stars = datamodel.Particles(number_of_stars)
        stars.id = id
        stars.mass = mass | nbody_system.mass
        stars.position = pos | nbody_system.length
        stars.velocity = vel | nbody_system.speed
        stars.radius = 0. | nbody_system.length

    # print "IDs:", stars.id.number
    sys.stdout.flush()

    #-----------------------------------------------------------------

    if softening_length == -1 | nbody_system.length:
        eps2 = 0.25*(float(number_of_stars))**(-0.666667) \
			| nbody_system.length**2
    else:
        eps2 = softening_length*softening_length

    gravity.parameters.timestep_parameter = accuracy_parameter
    gravity.parameters.epsilon_squared = eps2
    gravity.parameters.use_gpu = use_gpu
    # gravity.parameters.manage_encounters = manage_encounters

    print("adding particles")
    # print stars
    sys.stdout.flush()
    gravity.particles.add_particles(stars)
    gravity.commit_particles()

    print('')
    print("number_of_stars =", number_of_stars)
    print("evolving to time =", end_time.number, \
          "in steps of", delta_t.number)
    sys.stdout.flush()

    E0 = print_log(time, gravity)
    
    # Channel to copy values from the code to the set in memory.
    channel = gravity.particles.new_channel_to(stars)

    stopping_condition = gravity.stopping_conditions.collision_detection
    stopping_condition.enable()

    kep = Kepler(redirection = "none")
    kep.initialize_code()
    multiples_code = multiples.Multiples(gravity, new_smalln, kep)
    

    while time < end_time:
        time += delta_t
        multiples_code.evolve_model(time)

        # Copy values from the module to the set in memory.

        channel.copy()
    
        # Copy the index (ID) as used in the module to the id field in
        # memory.  The index is not copied by default, as different
        # codes may have different indices for the same particle and
        # we don't want to overwrite silently.

        channel.copy_attribute("index_in_code", "id")

        print_log(time, gravity, E0)
        sys.stdout.flush()

    print('')
    gravity.stop()

if __name__ == '__main__':

    infile = None
    N = 100
    t_end = 5.0 | nbody_system.time
    delta_t = 1.0 | nbody_system.time
    n_workers = 2
    use_gpu = 1
    gpu_worker = 1
    accuracy_parameter = 0.1
    softening_length = -1  | nbody_system.length
    random_seed = -1
    manage_encounters = 1

    try:
        opts, args = getopt.getopt(sys.argv[1:], "a:c:d:e:f:gGn:s:t:w:")
    except getopt.GetoptError as err:
        print(str(err))
        sys.exit(1)

    for o, a in opts:
        if o == "-a":
            accuracy_parameter = float(a)
        elif o == "-c":
            manage_encounters = int(a)
        elif o == "-d":
            delta_t = float(a) | nbody_system.time 
        elif o == "-e":
            softening_length = float(a) | nbody_system.length
        elif o == "-f":
            infile = a
        elif o == "-g":
            use_gpu = 0
        elif o == "-G":
            use_gpu = 0
            gpu_worker = 0
        elif o == "-n":
            N = int(a)
        elif o == "-s":
            random_seed = int(a)
        elif o == "-t":
            t_end = float(a) | nbody_system.time
        elif o == "-w":
            n_workers = int(a)
        else:
            print("unexpected argument", o)

    assert is_mpd_running()
    run_ph4(infile, N, t_end, delta_t, n_workers,
             use_gpu, gpu_worker,
             accuracy_parameter, softening_length,
             manage_encounters, random_seed)
